/*
    ggcap - ggcap.h
    Copyright (c) 2009-2010 Arkadiusz Bokowy
*/

#define APP_NAME "ggcap"
#define APP_VER "0.0.11"

#define __GGCAP_H

#include <unistd.h>
#include <stdlib.h>
#include <stdio.h>
#include <glib.h>
#include <time.h>
#include <pcap.h>
#include <libgadu.h>
#include "libgg80.h"

#include <libnet.h>
#include <netinet/if_ether.h>
#include <netinet/ip.h>
#include <netinet/udp.h>
#include <netinet/tcp.h>

void debug(char *msg);

#define GGCAP_MORALITY_NOTE

#define GG_NET     "91.214.237.0"
#define GG_NETMASK "255.255.255.128"

// UIN aliases table structure
#define GGCAP_MAX_ALIAS_SIZE 20
struct aliases {
	int uin;
	char alias[GGCAP_MAX_ALIAS_SIZE + 1];
};

// connection information structure
struct client2server_info {
	int type;
	time_t last_seen;
	uint32_t tcp_seq;  //for DUP detection
	uint32_t cliaddr, srvaddr;
	uint16_t cliprt, srvprt;
	uint8_t climac[ETH_ALEN], srvmac[ETH_ALEN];
};

// collected tcp-ip packet information structure
struct tcpip_pack_info {
	uint32_t saddr, daddr;
	uint16_t sprt, dprt;
	int pack_len, pack_type;
	int data_len;
	u_char *data;
};

#define SPOOFTOOL_C2S 1
#define SPOOFTOOL_S2C 2
struct spoof_data {
	libnet_t *lnet;

	libnet_ptag_t tcp, ipv4, ether;

	struct client2server_info *conn_info;
	int spoof_type;
};

// main application data structure
struct ggcap_data {
	char iface[IFNAMSIZ];
	uint32_t localip;
	pcap_t *pcap;

	char read_fname[100];

	uint32_t net, netmask;

	int uinal_size;
	struct aliases *uinal;

	GArray *cl2srv_db;
	GArray *tcpip_collector;

	// if 1 - redirect input data to spoof_input_callback
	int spooftool_input;

	int off_visual_packet_counter;

	struct flags {
		int main_loop:1, term_ok:1;
		int visualization:1, verbose:1;
		int drop_status:1, drop_local:1;
		int show_conn_onexit:1;
	}fl;
};

// global variables
struct ggcap_data gg_dt;
struct spoof_data sp_dt;

// in "snifftool.c"
void sniftool_dump_notifyreply(const u_char *pack, int len, int type);
void sniftool_dump_notify(const u_char *pack, int len, int type);
void sniftool_dump_login(const u_char *pack, int len, int type);
void sniftool_dump_welcome(const u_char *pack);
void sniftool_dump_status(const u_char *pack, int len, int type);
void sniftool_dump_newstatus(const u_char *pack, int len, int type);
void sniftool_dump_recvmsg(const u_char *pack, int len, int type);
void sniftool_dump_sendmsg(const u_char *pack, int len, int type);
char *sniftool_get_strstatus(int status);
char *sniftool_get_strpacktype(int to_server, int pack_type);
int sniftool_get_packet_type(const u_char *ip_pack, u_char *data, int *len);
int sniftool_is_gg_packet(const u_char *packet);

// in "spooftool.c"
int spooftool_input_callback();
void spooftools_disconnect_from_srv();
int spooftool_send_to_client(const u_char *data, int len);
int spooftool_send_to_server(const u_char *data, int len);
int spooftool_input_init(int init_char);
void spooftool_print_initial_info();
int spooftool_get_connection_no();
void spooftool_clean_up();
void spooftool_initialize();

// in "traficdb.c"
void traficdb_dumpdb();
int traficdb_updatedb(int to_server, int pack_type, const struct ether_header *eth_h,
	const struct iphdr *ip_h, const struct udphdr *tcp_udp_h, uint32_t tcp_seq);
int traficdb_get_last_seq(const struct client2server_info *cl2srv_info);
void traficdb_client_server_detect(struct client2server_info *cl2srv_inf,
	int to_server, const struct ether_header *eth_h, const struct iphdr *ip_h,
	const struct udphdr *tcp_udp_h);
int traficdb_get_client_type(int pack_type);
void traficdb_clean_up();
void traficdb_initialize();

// in "tcpipdata.c"
int tcpipdata_add_partial_data(const struct iphdr *ip_h,
	const struct udphdr *tcp_udp_h, u_char *data, int *len, int *type);
void tcpipdata_new_partial(const struct iphdr *ip_h,
	const struct udphdr *tcp_udp_h, int pack_type, int pack_len);
int tcpipdata_is_partial(const struct iphdr *ip_h,
	const struct udphdr *tcp_udp_h);
void tcpipdata_collector_clean_up();
void tcpipdata_collector_initialize();

// in "main.c"
void callback(u_char *args, const struct pcap_pkthdr *header,
	const u_char *packet);
void toggle_visualization_mode();
char *cp1250_to_utf8_plchar(char *msg);
char *get_uinalias(int uin);
int load_uinalias_table(const char *fname);
uint32_t get_devip(const char *iface);
char *strdumphex(void *mem, int len, char *buf);
void clean_up();
void exit_error(const char *errmsg);
void usage_help();
