/*
    ggcap - snifftool.c
    Copyright (c) 2009-2010 Arkadiusz Bokowy
*/

#include <string.h>
#include <errno.h>

#include <arpa/inet.h>
#include <netinet/ip.h>
#include <netinet/tcp.h>

#include "ggcap.h"

// Parse NOTIFY_REPLY packet and dump it to stdout
void sniftool_dump_notifyreply(const u_char *pack, int len, int type)
{
	int uin, descrlen, status, descr_data_seek, data_left = len;
	uint32_t cip = 0;
	uint16_t cport = 0;
	uint8_t ver = 0;
	struct gg_notify_reply *gg_nr;
	struct gg_notify_reply60 *gg_nr60;
	struct gg_notify_reply77 *gg_nr77;
	struct gg_status80 *gg_nr80; //gg_status80 is the same as gg_notify_reply80
	char *cp, descr[GG_STATUS_DESCR_MAXSIZE + 1];
	char *alias, alias_chr[50];
	char deb_str[200];

	// NOTE: this struct is append to gg_notify_reply if it contain
	//       description (for 60 and 77)
	struct gg_notify_reply_descr {
		uint8_t description_size;
//		char description[];     //NOT NULL-terminated
	} *gg_nrd;

	gg_nr = (struct gg_notify_reply *)(pack);
	gg_nr60 = (struct gg_notify_reply60 *)(pack);
	gg_nr77 = (struct gg_notify_reply77 *)(pack);
	gg_nr80 = (struct gg_status80 *)(pack);

	printf("  NOTIFY_REPLY:\n");

redo:
	// uin is first field in all structs
	uin = gg_nr->uin;

	switch(type){
	case GG_NOTIFY_REPLY80:
		status = gg_nr80->status;
		cip = gg_nr80->remote_ip;
		cport = gg_nr80->remote_port;
		descrlen = gg_nr80->description_size;
		cp = (char *)(gg_nr80 + 1);
printf("debug: diflen:%d", len-sizeof(gg_nr80));
		data_left=0;
		break;
	case GG_NOTIFY_REPLY77:
	case GG_NOTIFY_REPLY60:
		status = gg_nr60->status;
		cip = gg_nr60->remote_ip;
		cport = gg_nr60->remote_port;
		ver = gg_nr60->version;

		// parse description only if description flag is set
		if(GG_S_D(status)){
			if(type == GG_NOTIFY_REPLY77) //fix diff in struct size
				gg_nrd = (struct gg_notify_reply_descr *)(gg_nr77 + 1);
			else gg_nrd = (struct gg_notify_reply_descr *)(gg_nr60 + 1);

			descrlen = gg_nrd->description_size;
			cp = (char *)(gg_nrd + 1);
			descr_data_seek = sizeof(struct gg_notify_reply_descr) + descrlen;}
		else descr_data_seek = descrlen = 0;

		// set pointers to next struct for aproprietry client version
		if(type == GG_NOTIFY_REPLY77){
			gg_nr = gg_nr60 = gg_nr77 = &((char *)(gg_nr77 + 1))[descr_data_seek];
			data_left -= sizeof(struct gg_notify_reply77) + descr_data_seek;}
		else{
			gg_nr = gg_nr60 = gg_nr77 = &((char *)(gg_nr60 + 1))[descr_data_seek];
			data_left -= sizeof(struct gg_notify_reply60) + descr_data_seek;}
		break;
	case GG_NOTIFY_REPLY:
		status = gg_nr->status;
		cip = gg_nr->remote_ip;
		cport = gg_nr->remote_port;
		ver = gg_nr->version;
		descrlen = 0;

printf("debug: dun2:%x diflen:%d", gg_nr->dunno2, len-sizeof(gg_nr));
printf(" dump: %s\n", strdumphex(gg_nr++, 10, deb_str));
		data_left=0;
	}

	// copy description to our buffer (it isn't NULL terminated in packet)
	bzero(descr, sizeof(descr));
	memcpy(descr, cp, descrlen < sizeof(descr) ? descrlen : sizeof(descr));

	if((alias = get_uinalias(uin)) != NULL)
		sprintf(alias_chr, " (%s)", alias);
	else alias_chr[0] = 0;

	printf("  uin: %u%s -> %s", uin, alias_chr, sniftool_get_strstatus(status));
	printf(" || client IP: %s:%u & version: 0x%.2x\n",
			inet_ntoa(*((struct in_addr *)(&cip))), cport, ver);

	// print description only if status says that we've got one
	if(GG_S_D(status) || GG_S80_D(status))
		printf("    [%s]\n", cp1250_to_utf8_plchar(descr));

	if(data_left > 0) goto redo;
}

// Parse NOTIFY packet and dump it to stdout
void sniftool_dump_notify(const u_char *pack, int len, int type)
{
	int x;
	struct gg_notify *gg_n;

	gg_n = (struct gg_notify *)(pack);

	if(type == GG_ADD_NOTIFY) printf("  ADD USER:");
	else if(type == GG_REMOVE_NOTIFY) printf("  REMOVE USER:");
	else //userlist check at login (GG_NOTIFY_FIRST or GG_NOTIFY_LAST)
		printf("  USERLIST {");

	// dump all notify structs (mostly for GG_NOTIFY_FIRST and GG_NOTIFY_LAST)
	for(x = 0; x < len/sizeof(struct gg_notify); x++, gg_n++)
		printf(" (%u|%c%c%c)", gg_n->uin,
				(gg_n->dunno1 & GG_USER_BUDDY) == GG_USER_BUDDY ? 'l' : ' ',
				(gg_n->dunno1 & GG_USER_FRIEND) == GG_USER_FRIEND ? 'f' : ' ',
				(gg_n->dunno1 & GG_USER_BLOCKED) == GG_USER_BLOCKED ? 'b' : ' ');

	if(type == GG_ADD_NOTIFY || type == GG_REMOVE_NOTIFY) printf("\n");
	else //userlist check at login
		printf(" } %u entries\n", x);
}

// Parse LOGIN packet and dump it to stdout
void sniftool_dump_login(const u_char *pack, int len, int type)
{
	int uin, lip = 0, eip = 0;
	uint16_t lport = 0, eport = 0;
	uint8_t hash_tp, hash[64];
	struct gg_login_ext *gg_lx;
	struct gg_login60 *gg_l6;
	struct gg_login70 *gg_l7;
	struct gg_login80 *gg_l8;
	char *alias, alias_chr[50], hash_str[64*2+1];

	gg_lx = (struct gg_login_ext *)(pack);
	gg_l6 = (struct gg_login60 *)(pack);
	gg_l7 = (struct gg_login70 *)(pack);
	gg_l8 = (struct gg_login80 *)(pack);

	uin = gg_lx->uin; //in all structs it is first

	if((alias = get_uinalias(uin)) != NULL)
		sprintf(alias_chr, " (%s)", alias);
	else alias_chr[0] = 0;

	switch(type){
	case GG_LOGIN80:
		hash_tp = gg_l8->hash_type;
		memcpy(hash, &gg_l8->hash, sizeof(gg_l8->hash));
		lip = gg_l8->local_ip;
		lport = gg_l8->local_port;
		eip = gg_l8->external_ip;
		eport = gg_l8->external_port;
		break;
	case GG_LOGIN70:
		hash_tp = gg_l7->hash_type;
		memcpy(hash, &gg_l7->hash, sizeof(gg_l7->hash));
		lip = gg_l7->local_ip;
		lport = gg_l7->local_port;
		eip = gg_l7->external_ip;
		eport = gg_l7->external_port;
		break;
	case GG_LOGIN60:
		hash_tp = GG_LOGIN_HASH_GG32;
		memcpy(hash, &gg_l6->hash, sizeof(gg_l6->hash));
		lip = gg_l6->local_ip;
		lport = gg_l6->local_port;
		eip = gg_l6->external_ip;
		eport = gg_l6->external_port;
		break;
	case GG_LOGIN_EXT:
		eip = gg_lx->external_ip;
		eport = gg_lx->external_port;
	case GG_LOGIN:
		hash_tp = GG_LOGIN_HASH_GG32;
		memcpy(hash, &gg_lx->hash, sizeof(gg_lx->hash));
		lip = gg_lx->local_ip;
		lport = gg_lx->local_port;
	}

//TODO: cos nie tak z tym hashem araz z IP:port dla gg80 (reszta nie testowana)

	printf("  LOGIN (uin: %u%s) [%s]\n", uin, alias_chr,
			hash_tp == GG_LOGIN_HASH_SHA1 ? "SHA1": "GG32");
	printf("  hash: [%s]\n", strdumphex(hash, sizeof(hash), hash_str));
	printf("  local IP: %s:%u", inet_ntoa(*((struct in_addr *)(&lip))), lport);
	printf(" || external IP: %s:%u\n",
			inet_ntoa(*((struct in_addr *)(&eip))), eport);
}

// Parse WELCOME packet and dump it to stdout
void sniftool_dump_welcome(const u_char *pack)
{
	struct gg_welcome *gg_w;

	gg_w = (struct gg_welcome *)(pack);
	printf("  WELCOME (password encryption key: 0x%.4x)\n", gg_w->key);
}

// Parse STATUS packet and dump it to stdout
void sniftool_dump_status(const u_char *pack, int len, int type)
{
	int uin, descrlen, status;
	uint32_t cip = 0;
	uint16_t cport = 0;
	uint8_t ver = 0;
	struct gg_status *gg_s;
	struct gg_status60 *gg_s60;
	struct gg_status77 *gg_s77;
	struct gg_status80 *gg_s80;
	char *cp, descr[GG_STATUS_DESCR_MAXSIZE + 1];
	char *alias, alias_chr[50];

	gg_s = (struct gg_status *)(pack);
	gg_s60 = (struct gg_status60 *)(pack);
	gg_s77 = (struct gg_status77 *)(pack);
	gg_s80 = (struct gg_status80 *)(pack);

	// uin is first field in all structs
	uin = gg_s->uin;

	switch(type) {
	case GG_STATUS80:
		status = gg_s80->status;
		cip = gg_s80->remote_ip;
		cport = gg_s80->remote_port;
		descrlen = gg_s80->description_size;
		cp = (char *)(gg_s80 + 1);
		break;
	case GG_STATUS77: //there is small difference between stat60 and stat77, so
	case GG_STATUS60: //we can write it in this way
		status = gg_s60->status;
		cip = gg_s60->remote_ip;
		cport = gg_s60->remote_port;
		ver = gg_s60->version;

		if(type == GG_STATUS77){
			descrlen = len - sizeof(struct gg_status77);
			cp = (char *)(gg_s77 + 1);}
		else{
			descrlen = len - sizeof(struct gg_status60);
			cp = (char *)(gg_s60 + 1);}
		break;
	case GG_STATUS:
		status = gg_s->status;
		descrlen = len - sizeof(struct gg_status);
		cp = (char *)(gg_s + 1);
	}

	// copy description to our buffer (it isn't NULL terminated in packet)
	bzero(descr, sizeof(descr));
	memcpy(descr, cp, descrlen < sizeof(descr) ? descrlen : sizeof(descr));

	if((alias = get_uinalias(uin)) != NULL)
		sprintf(alias_chr, " (%s)", alias);
	else alias_chr[0] = 0;

	printf("  STATUS (uin: %u%s -> %s)\n", uin, alias_chr,
			sniftool_get_strstatus(status));
	printf("  client IP: %s:%u || client version: 0x%.2x\n",
			inet_ntoa(*((struct in_addr *)(&cip))), cport, ver);

	// print description only if status says that we've got one
	if(GG_S_D(status) || GG_S80_D(status))
		printf("  [%s]\n", cp1250_to_utf8_plchar(descr));
}

// Parse NEW_STATUS packet and dump it to stdout
void sniftool_dump_newstatus(const u_char *pack, int len, int type)
{
	int descrlen;
	struct gg_new_status *gg_ns;
	struct gg_new_status80 *gg_ns80;
	char *cp, descr[GG_STATUS_DESCR_MAXSIZE + 1];

	gg_ns = (struct gg_new_status *)(pack);
	gg_ns80 = (struct gg_new_status80 *)(pack);

	if(type == GG_NEW_STATUS80){
		descrlen = gg_ns80->description_size;
		cp = (char *)(gg_ns80 + 1);}
	else{
		descrlen = len - sizeof(struct gg_new_status);
		cp = (char *)(gg_ns + 1);}

	// copy description to our buffer (it isn't NULL terminated in packet)
	bzero(descr, sizeof(descr));
	memcpy(descr, cp, descrlen < sizeof(descr) ? descrlen : sizeof(descr));

	printf("  NEW_STATUS (%s)\n", sniftool_get_strstatus(gg_ns->status));

	// print description only if status says that we've got one
	if(GG_S_D(gg_ns->status) || GG_S80_D(gg_ns->status))
		printf("  [%s]\n", cp1250_to_utf8_plchar(descr));
}

// Parse RECV_MSG packet and dump it to stdout
void sniftool_dump_recvmsg(const u_char *pack, int len, int type)
{
	time_t tm_t;
	struct tm tm_s;
	struct gg_recv_msg *gg_rm;
	struct gg_recv_msg80 *gg_rm80;
	char *cp, *alias, alias_chr[50];

	gg_rm = (struct gg_recv_msg *)(pack);
	gg_rm80 = (struct gg_recv_msg80 *)(pack);

	tm_t = gg_rm->time;
	gmtime_r(&tm_t, &tm_s);

	if((alias = get_uinalias(gg_rm->sender)) != NULL)
		sprintf(alias_chr, " (%s)", alias);
	else alias_chr[0] = 0;

	if(type == GG_RECV_MSG80)
		cp = &((char *)(gg_rm80))[gg_rm80->offset_plain];
	else cp = (char *)(gg_rm + 1);

	// header is the same in all GG protocol versions
	printf("  RECV_MSG < %u%s [time: %.2d:%.2d.%.2d] (msgclass: 0x%.2x)\n",
			gg_rm->sender, alias_chr, tm_s.tm_hour, tm_s.tm_min, tm_s.tm_sec,
			gg_rm->msgclass);

	printf("  [%s]\n", cp1250_to_utf8_plchar(cp));
}

// Parse SEND_MSG packet and dump it to stdout
void sniftool_dump_sendmsg(const u_char *pack, int len, int type)
{
	struct gg_send_msg *gg_sm;
	struct gg_send_msg80 *gg_sm80;
	char *cp, *alias, alias_chr[50];

	gg_sm = (struct gg_send_msg *)(pack);
	gg_sm80 = (struct gg_send_msg80 *)(pack);

	if((alias = get_uinalias(gg_sm->recipient)) != NULL)
		sprintf(alias_chr, " (%s)", alias);
	else alias_chr[0] = 0;

	if(type == GG_SEND_MSG80)
		cp = &((char *)(gg_sm80))[gg_sm80->offset_plain];
	else cp = (char *)(gg_sm + 1);

	// header is the same in all GG protocol versions
	printf("  SEND_MSG > %u%s (msgclass: 0x%.2x)\n",
			gg_sm->recipient, alias_chr, gg_sm->msgclass);

	printf("  [%s]\n", cp1250_to_utf8_plchar(cp));
}

// Return human readable user status information
char *sniftool_get_strstatus(int status)
{
	int friends_only = 0, graph_descr = 0;
	char status_txt[][10] = {"un: 0x00", "online", "offline",
			"busy", "invisible", "block"};
	char *cp, ffo[] = " | for friends only", gd[] = " | graphical description";
	static char strbuff[64];

	// detect and remove all flags
	if(GG_S_F(status)) friends_only = 1;
	if(GG_S80_GD(status)) graph_descr = 1;

	if(GG_S_A(GG_S80(status))) cp = status_txt[1];
	else if(GG_S_NA(GG_S80(status))) cp = status_txt[2];
	else if(GG_S_B(GG_S80(status))) cp = status_txt[3];
	else if(GG_S_I(GG_S80(status))) cp = status_txt[4];
	else if(GG_S_BL(GG_S80(status))) cp = status_txt[5];
	else{
		cp = status_txt[0];
		sprintf(cp, "un: 0x%.2x", (char)(status));}

	sprintf(strbuff, "%s%s%s", cp, friends_only ? ffo : "", graph_descr ? gd : "");
	return strbuff;
}

// Return human readable gg packet_type string
char *sniftool_get_strpacktype(int to_server, int pack_type)
{
	static char *strintype[] = {"WELCOME", "STATUS", "LOGIN_OK", NULL,
"SEND_MSG_ACK", NULL, "PONG", "PING", "LOGIN_FAILED", "RECV_MSG",
"DISCONNECTING", "NOTIFY_REPLY", "DISCONNECT_ACK", "PUBDIR50_REPLY",
"STATUS60", "USERLIST_REPLY", "NOTIFY_REPLY60", NULL, NULL, "NEED_EMAIL",
NULL, "LOGIN_HASH_TYPE_INVALID", "STATUS77", "NOTIFY_REPLY77", NULL, NULL,
NULL, NULL, NULL, NULL, "DCC7_INFO", "DCC7_NEW", "DCC7_ACCEPT", "DCC7_REJECT",
"DCC7_ID_REPLY", NULL, "DCC7_ABORTED", NULL, "XML_EVENT", NULL, NULL, NULL,
"NOTIFY_REPLY80BETA", "XML_ACTION", NULL, "RECV_MSG80", NULL, "USERLIST_REPLY80",
NULL, NULL, NULL, NULL, "LOGIN80_OK", "STATUS80", "NOTIFY_REPLY80"};
	static char *strouttype[] = {NULL, "NEW_STATUS", NULL, NULL, NULL, NULL,
"PONG", "PING", NULL, NULL, "SEND_MSG", "LOGIN", "ADD_NOTIFY", "REMOVE_NOTIFY",
"NOTIFY_FIRST", "NOTIFY_LAST", NULL, "LIST_EMPTY", "LOGIN_EXT", "PUBDIR50_REQUEST",
"LOGIN60", "USERLIST_REQUEST", NULL, NULL, "LOGIN70", NULL, NULL, NULL, NULL,
NULL, "DCC7_INFO", "DCC7_NEW", "DCC7_ACCEPT", "DCC7_REJECT", "DCC7_ID_REQUEST",
"DCC7_DUNNO1", "DCC7_ABORT", NULL, NULL, "NEW_STATUS80BETA", "LOGIN80BETA",
NULL, NULL, NULL, "SEND_MSG80", NULL, "USERLIST_REQUEST80", NULL, "LOGIN80",
NULL, NULL, NULL, NULL, NULL, NULL, "NEW_STATUS80"};
	// NOTE: if you increase size of this tables you MUST update upper limit
	//       pack_type protection!

	// low limit segfault protection
	if(pack_type < 1) return NULL;

	if(to_server){
		if(pack_type > 56) return NULL; //upper limit for strouttype
		return strouttype[pack_type - 1];}

	if(pack_type > 55) return NULL; //upper limit for strintype
	return strintype[pack_type - 1];
}

// Return GG packet type (return -1 if partial packet data)
int sniftool_get_packet_type(const u_char *ip_pack, u_char *data, int *len)
{
	int tcp_ip_data_size;
	struct iphdr *ip_h;
	struct tcphdr *tcp_h;
	struct gg_header *gg_h;

	ip_h = (struct iphdr *)(ip_pack);
	tcp_h = (struct tcphdr *)(ip_h + 1);

	// NOTE: tcphdr_size = tcp_h->doff*(32 bites word)
	gg_h = (struct gg_header *)((char *)tcp_h + 4*tcp_h->doff);

	tcp_ip_data_size = ntohs(ip_h->tot_len) - (int)((char *)gg_h - (char *)ip_h);

	// return 0 if there is no data in tcp_ip stack
	if(tcp_ip_data_size == 0) return 0;

	// if true we've got next part of partial data
	if(tcpipdata_is_partial(ip_h, (struct udphdr *)tcp_h) != -1){
		// dont remove gg_header because there is no such header
		*len = tcp_ip_data_size;
		memcpy(data, (void *)(gg_h), *len);
		return -1;}

	// return data encapsulated in gg_h without this header
	*len = tcp_ip_data_size - sizeof(struct gg_header);
	memcpy(data, (void *)(gg_h + 1), *len);

	// if we've got less data in tcp than gg_h needs start new partial packet
	// NOTE: Adding data to collector is in main.c after DUP packets detection!
	if(*len < gg_h->length){
		// NOTE: if gg_h->len is > available data but gg_h->type is unrecognizable
		//       it could mean that this packet is DUP of LAST part of fragmented
		//       data (tcpipdata_is_partial did not detect fragmentation
		//       because this module already release this fragment data info)
		if(sniftool_get_strpacktype(0, gg_h->type) == NULL &&
				sniftool_get_strpacktype(1, gg_h->type) == NULL)
			return 0;

		tcpipdata_new_partial(ip_h, (struct udphdr *)tcp_h, gg_h->type, gg_h->length);
		return -1;}

	return gg_h->type;
}

// Test if packet is Gadu-Gadu protocol packet
// NOTE: retval bit info:
//   1 - default port
//   2 - outgoing
//   3 - incoming (server OK)
//   4 - alternative port (HTTPS)
//   5 - HTTP (login process)
int sniftool_is_gg_packet(const u_char *packet)
{
	int server = 0;
	uint16_t psrc, pdst;
 	struct ether_header *eth_h;
	struct iphdr *ip_h;
	struct tcphdr *tcp_h;

	eth_h = (struct ether_header *)(packet);
	ip_h = (struct iphdr *)(eth_h + 1);
	tcp_h = (struct tcphdr *)(ip_h + 1); //FIXME: what if IP has options!!???

	// gg packets are only TCP/IP packets
	if(eth_h->ether_type != 8) return 0;
	if(ip_h->protocol != 6) return 0;

	psrc = ntohs(tcp_h->source);
	pdst = ntohs(tcp_h->dest);

	// check all known Gadu-Gadu servers
	if((ip_h->daddr & gg_dt.netmask) == gg_dt.net) server = 0x02;
	if((ip_h->saddr & gg_dt.netmask) == gg_dt.net) server = 0x04;

	// default port matches so... it schould be a gg packet
	// NOTE: server maching check skipped
	if(psrc == GG_DEFAULT_PORT || pdst == GG_DEFAULT_PORT)
		return 0x01 | server;

	// last chance port 443 (HTTPS)
	if(server && (psrc == GG_HTTPS_PORT || pdst == GG_HTTPS_PORT))
		return 0x08 | server;

	// HTTP port (login process ;>)
	if(server && (psrc == 80 || pdst == 80))
		return 0x16 | server;

	return 0;
}
