/*
    ggcap - tcpipdata.c
    Copyright (c) 2009 Arkadiusz Bokowy
*/

#include <string.h>

#include "ggcap.h"

// Add data to collected packet or open new collect thread.
// NOTE: If this function returns 1 it means that packet is ready. Returned
//       gg packet (without header) is in data and its len is in len,
//       in type gg packet_type is returned.
int tcpipdata_add_partial_data(const struct iphdr *ip_h,
	const struct udphdr *tcp_udp_h, u_char *data, int *len, int *type)
{
	struct tcpip_pack_info *db_inf;
	int index;

	index = tcpipdata_is_partial(ip_h, tcp_udp_h);
	if(index == -1) return -1;

	db_inf = &g_array_index(gg_dt.tcpip_collector, struct tcpip_pack_info, index);

	// overflow protection (just in case)
	if(db_inf->data_len + *len > db_inf->pack_len)
		*len = db_inf->pack_len - db_inf->data_len;

	memcpy(&db_inf->data[db_inf->data_len], data, *len);
	db_inf->data_len += *len;

	// packet is completed -> give it back :)
	if(db_inf->data_len >= db_inf->pack_len){
		memcpy(data, db_inf->data, db_inf->pack_len);
		*len = db_inf->pack_len;
		*type = db_inf->pack_type;

		// free allocated memory
		free(db_inf->data);
		g_array_remove_index_fast(gg_dt.tcpip_collector, index);
		return 1;}

	return 0;
}

// Add new packet collection info to db (allocate memory)
void tcpipdata_new_partial(const struct iphdr *ip_h,
	const struct udphdr *tcp_udp_h, int pack_type, int pack_len)
{
	struct tcpip_pack_info tcpip_inf;

	bzero(&tcpip_inf, sizeof(tcpip_inf));

	// Prepare initial info struct
	tcpip_inf.saddr = ip_h->saddr;
	tcpip_inf.daddr = ip_h->daddr;
	tcpip_inf.sprt = tcp_udp_h->source;
	tcpip_inf.dprt = tcp_udp_h->dest;
	tcpip_inf.pack_type = pack_type;
	tcpip_inf.pack_len = pack_len;

	// alloc memory for pasket data
	tcpip_inf.data = malloc(pack_len);

	// not enough memory for collecting packet... :/ maybe pack_len is too long
	// what could mean that data were corrupted - just skip collection
	if(tcpip_inf.data == NULL) return;

	g_array_append_val(gg_dt.tcpip_collector, tcpip_inf);
}

// Return tcpip_pack_info struct index if this packet is part of collected
// one or -1 if not. If this function returns -2 it means that we've got
// DUP packet.
int tcpipdata_is_partial(const struct iphdr *ip_h,
	const struct udphdr *tcp_udp_h)
{
	struct tcpip_pack_info *db_inf;
	int x;

	for(x = 0; x < gg_dt.tcpip_collector->len; x++){
		db_inf = &g_array_index(gg_dt.tcpip_collector, struct tcpip_pack_info, x);

		// if src and dest ip:port match indicate partial
		if(db_inf->saddr == ip_h->saddr && db_inf->daddr == ip_h->daddr &&
				db_inf->sprt == tcp_udp_h->source && db_inf->dprt == tcp_udp_h->dest)
			return x;
	}

	return -1;
}

void tcpipdata_collector_clean_up()
{
	struct tcpip_pack_info *db_inf;
	int x;

	for(x = 0; x < gg_dt.tcpip_collector->len; x++){
		db_inf = &g_array_index(gg_dt.tcpip_collector, struct tcpip_pack_info, x);
		free(db_inf->data);}

	// destroy database array (free elements too)
	if(gg_dt.tcpip_collector) g_array_free(gg_dt.tcpip_collector, TRUE);
}

void tcpipdata_collector_initialize()
{
	// create database array (preload 5 entry)
	gg_dt.tcpip_collector = g_array_sized_new(0, 0,
			sizeof(struct tcpip_pack_info), 5);
}
